@@ -3,10 +3,10 @@
  *
  * Author:	Miquel van Smoorenburg, miquels@cistron.nl
  * 
- * Version:	@(#)dowall.c  2.76  19-Apr-1999  miquels@cistron.nl
+ * Version:	@(#)dowall.c  2.79  11-Sep-2000  miquels@cistron.nl
  *
  *		This file is part of the sysvinit suite,
- *		Copyright 1991-1999 Miquel van Smoorenburg.
+ *		Copyright 1991-2000 Miquel van Smoorenburg.
  *
  *		This program is free software; you can redistribute it and/or
  *		modify it under the terms of the GNU General Public License
@@ -28,7 +28,8 @@
 
 static jmp_buf jbuf;
 
-#define AEROSMITH
+/* display pid of rpc.rwalld if wall() is called with 'remote' non-zero */
+#define RWALLDPID
 
 /*
  *	Alarm handler
@@ -68,29 +69,32 @@ void wall(char *text, int fromshutdown, int remote)
 {
 	FILE *tp;
 	char line[81];
-	char term[32];
-	static char *user, ttynm[16], *date;
+	char term[UT_LINESIZE+6];
+	static char *user, ttynm[32], *date;
+	char uidbuf[32];
 	static int fd, init = 0;
 	struct passwd *pwd;
 	char *tty, *p;
 	time_t t;
 	struct utmp *utmp;
-	int uid;
-	
+	uid_t uid;
+
 	setutent();
-	
+
 	if (init == 0) {
 		uid = getuid();
-		if ((pwd = getpwuid(uid)) == (struct passwd *)0 &&
-		    (uid != 0 || fromshutdown)) {
-			fprintf(stderr, "You don't exist. Go away.\n");
-			exit(1);
-		}
-		user = pwd ? pwd->pw_name : "root";
-		if ((tty = ttyname(0)) != (char *)0) {
+		pwd = getpwuid(uid);
+		sprintf (uidbuf, "uid %d", (int)uid);
+		/*
+		 *	display uid if no corresponding username found in
+		 *	password file. superuser (uid 0) is always "root",
+		 *	not "uid 0".
+		 */
+		user = pwd ? pwd->pw_name : (uid == 0 ? "root" : uidbuf);
+		if ((tty = ttyname(0)) != NULL) {
 			if (strncmp(tty, "/dev/", 5) == 0)
 				tty += 5;
-			sprintf(ttynm, "(%s) ", tty);	
+			sprintf(ttynm, "(%.28s) ", tty);	
 		} else
 			ttynm[0] = 0;
 		init++;
@@ -105,12 +109,12 @@ void wall(char *text, int fromshutdown, int remote)
 	*p = 0;
 	
 	if (remote) {
-		sprintf(line,
-			"\007\r\nRemote broadcast message %s...\r\n\r\n",
+		snprintf(line, sizeof(line),
+			"\007\r\nRemote broadcast message (%s):\r\n\r\n",
 			date);
 	} else {
 		snprintf(line, sizeof(line),
-			"\007\r\nBroadcast message from %s %s%s...\r\n\r\n",
+			"\007\r\nBroadcast message from %s %s(%s):\r\n\r\n",
 			user, ttynm, date);
 	}
 
@@ -124,10 +128,12 @@ void wall(char *text, int fromshutdown, int remote)
 	while ((utmp = getutent()) != NULL) {
 		if(utmp->ut_type != USER_PROCESS ||
 		   utmp->ut_user[0] == 0) continue;
+		term[sizeof(term) - 1] = 0;
 		if (strncmp(utmp->ut_line, "/dev/", 5) == 0)
-			strcpy(term, utmp->ut_line);
+			strncpy(term, utmp->ut_line, sizeof(term)-1);
 		else
-			sprintf(term, "/dev/%s", utmp->ut_line);
+			snprintf(term, sizeof(term), "/dev/%s",
+				utmp->ut_line);
 	
 		/*
 		 *	Sometimes the open/write hangs in spite of the O_NDELAY
@@ -142,12 +148,6 @@ void wall(char *text, int fromshutdown, int remote)
 			if ((tp = fdopen(fd, "w")) != NULL) {
 				fputs(line, tp);
 				feputs(text, tp);
-#ifdef AEROSMITH
-				if (fromshutdown &&
-				    !strcmp(utmp->ut_user, "tyler"))
-				 fputs("Oh hello Mr. Tyler - going DOWN?\r\n",
-					tp);
-#endif
 				fclose(tp);
 			} else
 				close(fd);
