@@ -5,10 +5,10 @@
  *		     init [0123456SsQqAaBbCc]
  *		  telinit [0123456SsQqAaBbCc]
  *
- * Version:	@(#)init.c  2.83  05-Sep-2001  miquels@cistron.nl
+ * Version:	@(#)init.c  2.84  27-Nov-2001  miquels@cistron.nl
  */
-#define VERSION "2.83"
-#define DATE    "01-Oct-2001"
+#define VERSION "2.84"
+#define DATE    "27-Nov-2001"
 /*
  *		This file is part of the sysvinit suite,
  *		Copyright 1991-2001 Miquel van Smoorenburg.
@@ -705,8 +705,11 @@ void set_term(int how)
 
 	/*
 	 *	Now set the terminal line.
+	 *	We don't care about non-transmitted output data
+	 *	and non-read input data.
 	 */
-	(void) tcsetattr(fd, TCSADRAIN, &tty);
+	(void) tcsetattr(fd, TCSANOW, &tty);
+	(void) tcflush(fd, TCIOFLUSH);
 	(void) close(fd);
 }
 
@@ -1480,52 +1483,52 @@ void read_inittab(void)
  */
 void start_if_needed(void)
 {
-  CHILD *ch;		/* Pointer to child */
-  int delete;		/* Delete this entry from list? */
+	CHILD *ch;		/* Pointer to child */
+	int delete;		/* Delete this entry from list? */
 
 #if DEBUG
-  log(L_VB, "Checking for children to start");
+	log(L_VB, "Checking for children to start");
 #endif
 
-  for(ch = family; ch; ch = ch->next) {
+	for(ch = family; ch; ch = ch->next) {
 
 #if DEBUG
-	if (ch->rlevel[0] == 'C') {
-		log(L_VB, "%s: flags %d", ch->process, ch->flags);
-	}
+		if (ch->rlevel[0] == 'C') {
+			log(L_VB, "%s: flags %d", ch->process, ch->flags);
+		}
 #endif
 
-	/* Are we waiting for this process? Then quit here. */
-	if (ch->flags & WAITING) break;
-
-	/* Already running? OK, don't touch it */
-	if (ch->flags & RUNNING) continue;
-
-	/* See if we have to start it up */
-	delete = 1;
-	if (strchr(ch->rlevel, runlevel) ||
-	    ((ch->flags & DEMAND) && !strchr("#*Ss", runlevel))) {
-		startup(ch);
-		delete = 0;
-	}
-
-	if (delete) {
-		/* FIXME: is this OK? */
-		ch->flags &= ~(RUNNING|WAITING);
-		if (!ISPOWER(ch->action) && ch->action != KBREQUEST)
-			ch->flags &= ~XECUTED;
-		ch->pid = 0;
-	} else
-		/* Do we have to wait for this process? */
+		/* Are we waiting for this process? Then quit here. */
 		if (ch->flags & WAITING) break;
-  }
-  /* Done. */
+
+		/* Already running? OK, don't touch it */
+		if (ch->flags & RUNNING) continue;
+
+		/* See if we have to start it up */
+		delete = 1;
+		if (strchr(ch->rlevel, runlevel) ||
+		    ((ch->flags & DEMAND) && !strchr("#*Ss", runlevel))) {
+			startup(ch);
+			delete = 0;
+		}
+
+		if (delete) {
+			/* FIXME: is this OK? */
+			ch->flags &= ~(RUNNING|WAITING);
+			if (!ISPOWER(ch->action) && ch->action != KBREQUEST)
+				ch->flags &= ~XECUTED;
+			ch->pid = 0;
+		} else
+			/* Do we have to wait for this process? */
+			if (ch->flags & WAITING) break;
+	}
+	/* Done. */
 }
 
 /*
  *	Ask the user on the console for a runlevel
  */
-int ask_runlevel()
+int ask_runlevel(void)
 {
 	const char	prompt[] = "\nEnter runlevel: ";
 	char		buf[8];
@@ -1826,7 +1829,6 @@ void re_exec(void)
 
 	if (strchr("S12345",runlevel) == NULL)
 		return;
-	closelog();			/* AAAARRRGH! It _is_ needed */
 
 	/*
 	 *	Reset the alarm, and block all signals.
@@ -1868,7 +1870,7 @@ void re_exec(void)
 		/*
 		 *	Yup. _Parent_ exec's ...
 		 */
-		execl(myname, myname, NULL);
+		execl(myname, myname, "--init", NULL);
 	} else if (pid == 0) {
 		/*
 		 *	... while child sends her the
@@ -1894,34 +1896,35 @@ void re_exec(void)
  */
 void fifo_new_level(int level)
 {
-  int oldlevel;
+	int oldlevel;
 #if CHANGE_WAIT
-  CHILD *ch;
+	CHILD *ch;
 #endif
 
-  if (level == runlevel) return;
+	if (level == runlevel) return;
 
 #if CHANGE_WAIT
-  /* Are we waiting for a child? */
-  for(ch = family; ch; ch = ch->next)
-	if (ch->flags & WAITING) break;
-  if (ch == NULL)
+	/* Are we waiting for a child? */
+	for(ch = family; ch; ch = ch->next)
+		if (ch->flags & WAITING) break;
+	if (ch == NULL)
 #endif
-  {
-	/* We need to go into a new runlevel */
-	oldlevel = runlevel;
-	runlevel = read_level(level);
-	if (runlevel == 'U') {
-		runlevel = oldlevel;
-		re_exec();
-	} else {
-		if (oldlevel != 'S' && runlevel == 'S') set_term(0);
-		if (runlevel == '6' || runlevel == '0' || runlevel == '1') set_term(0);
-		read_inittab();
-		fail_cancel();
-		setproctitle("init [%c]", runlevel);
+	{
+		/* We need to go into a new runlevel */
+		oldlevel = runlevel;
+		runlevel = read_level(level);
+		if (runlevel == 'U') {
+			runlevel = oldlevel;
+			re_exec();
+		} else {
+			if (oldlevel != 'S' && runlevel == 'S') set_term(0);
+			if (runlevel == '6' || runlevel == '0' ||
+			    runlevel == '1') set_term(0);
+			read_inittab();
+			fail_cancel();
+			setproctitle("init [%c]", runlevel);
+		}
 	}
-  }
 }
 
 /*
@@ -2444,34 +2447,106 @@ void Usage(char *s)
   exit(1);
 }
 
+int telinit(char *progname, int argc, char **argv)
+{
+#ifdef TELINIT_USES_INITLVL
+	FILE			*fp;
+#endif
+	struct init_request	request;
+	struct sigaction	sa;
+	int			f, fd;
+
+	while((f = getopt(argc, argv, "t:")) != EOF) switch(f) {
+		case 't':
+			sltime = atoi(optarg);
+			break;
+		default:
+			Usage(progname);
+			break;
+	}
+
+	/* Check syntax. */
+	if (argc - optind != 1 || strlen(argv[optind]) != 1) Usage(progname);
+	if (!strchr("0123456789SsQqAaBbCcUu", argv[optind][0])) Usage(progname);
+
+	/* Open the fifo and write a command. */
+	memset(&request, 0, sizeof(request));
+	request.magic     = INIT_MAGIC;
+	request.cmd       = INIT_CMD_RUNLVL;
+	request.runlevel  = argv[optind][0];
+	request.sleeptime = sltime;
+
+	/* Make sure we don't hang on opening /dev/initctl */
+	SETSIG(sa, SIGALRM, signal_handler, 0);
+	alarm(3);
+	if ((fd = open(INIT_FIFO, O_WRONLY)) >= 0 &&
+	    write(fd, &request, sizeof(request)) == sizeof(request)) {
+		close(fd);
+		alarm(0);
+		return 0;
+	}
+#ifndef TELINIT_USES_INITLVL
+	fprintf(stderr, "%s: ", progname);
+	if (ISMEMBER(got_signals, SIGALRM)) {
+		fprintf(stderr, "timeout opening/writing control channel %s\n",
+			INIT_FIFO);
+	} else {
+		perror(INIT_FIFO);
+	}
+	return 1;
+#endif
+
+	/* Fallthrough to the old method. */
+
+#ifdef TELINIT_USES_INITLVL
+	/* Now write the new runlevel. */
+	if ((fp = fopen(INITLVL, "w")) == NULL) {
+		fprintf(stderr, "%s: cannot create %s\n", progname, INITLVL);
+		exit(1);
+	}
+	fprintf(fp, "%s %d", argv[optind], sltime);
+	fclose(fp);
+
+	/* And tell init about the pending runlevel change. */
+	if (kill(INITPID, SIGHUP) < 0) perror(progname);
+
+	return 0;
+#endif
+}
 
 /*
  * Main entry for init and telinit.
  */
-int main(int argc, char *argv[])
+int main(int argc, char **argv)
 {
-  struct init_request	request;
-  struct sigaction	sa;
-  FILE			*fp;
-  char			*p;
-  int			f, fd;
+	char			*p;
+	int			f;
+	int			isinit;
 
-  /* Get my own name */
-  if ((p = strrchr(argv[0], '/')) != NULL)
-  	p++;
-  else
-  	p = argv[0];
-  	
+	/* Get my own name */
+	if ((p = strrchr(argv[0], '/')) != NULL)
+  		p++;
+	else
+  		p = argv[0];
+	umask(022);
 
-  umask(022);
+	/* Quick check */
+	if (geteuid() != 0) {
+		fprintf(stderr, "%s: must be superuser.\n", p);
+		exit(1);
+	}
+
+	/*
+	 *	Is this telinit or init ?
+	 */
+	isinit = (getpid() == 1);
+	for (f = 1; f < argc; f++) {
+		if (!strcmp(argv[f], "-i") || !strcmp(argv[f], "--init"))
+			isinit = 1;
+			break;
+	}
+	if (!isinit) exit(telinit(p, argc, argv));
 
-  /*
-   *	Is this telinit or init ?
-   */
-#ifndef TEST
-  if (getpid() == INITPID || !strcmp(p, "init.new") || !strcmp(p, "sh"))
-#endif
-  {
 	/*
 	 *	Check for re-exec
 	 */ 	
@@ -2496,7 +2571,7 @@ int main(int argc, char *argv[])
 			dfl_level = 'S';
 		else if (!strcmp(argv[f], "-a") || !strcmp(argv[f], "auto"))
 			putenv("AUTOBOOT=YES");
-		else if (!strcmp(argv[f], "-b") || !strcmp(argv[f], "emergency"))
+		else if (!strcmp(argv[f], "-b") || !strcmp(argv[f],"emergency"))
 			emerg_shell = 1;
 		else if (!strcmp(argv[f], "-z")) {
 			/* Ignore -z xxx */
@@ -2515,63 +2590,7 @@ int main(int argc, char *argv[])
 	argv[1] = NULL;
 	setproctitle("init boot");
 	init_main(dfl_level);
-  }
 
-  /* Nope, this is a change-run-level init */
-  while((f = getopt(argc, argv, "t:")) != EOF) {
-	switch(f) {
-		case 't':
-			sltime = atoi(optarg);
-			break;
-		default:
-			Usage(p);
-			break;
-	}
-  }
-
-  if (geteuid() != 0) {
-	fprintf(stderr, "init: must be superuser.\n");
-	exit(1);
-  }
-
-  /* Check syntax. */
-  if (argc - optind != 1 || strlen(argv[optind]) != 1) Usage(p);
-  if (!strchr("0123456789SsQqAaBbCcUu", argv[optind][0])) Usage(p);
-
-  /* Open the fifo and write a command. */
-  memset(&request, 0, sizeof(request));
-  request.magic     = INIT_MAGIC;
-  request.cmd       = INIT_CMD_RUNLVL;
-  request.runlevel  = argv[optind][0];
-  request.sleeptime = sltime;
-
-  /* Make sure we don't hang on opening /etc/init.fifo */
-  SETSIG(sa, SIGALRM, signal_handler, 0);
-  alarm(3);
-  if ((fd = open(INIT_FIFO, O_WRONLY)) >= 0 &&
-	    write(fd, &request, sizeof(request)) == sizeof(request)) {
-	close(fd);
-	alarm(0);
+	/*NOTREACHED*/
 	return 0;
-  }
-#ifndef INITLVL
-  perror(INIT_FIFO);
-  exit(1);
-#endif
-  /* Fallthrough to the old method. */
-
-#ifdef INITLVL
-  /* Now write the new runlevel. */
-  if ((fp = fopen(INITLVL, "w")) == NULL) {
-	fprintf(stderr, "%s: cannot create %s\n", p, INITLVL);
-	exit(1);
-  }
-  fprintf(fp, "%s %d", argv[optind], sltime);
-  fclose(fp);
-
-  /* And tell init about the pending runlevel change. */
-  if (kill(INITPID, SIGHUP) < 0) perror(p);
-
-  exit(0);
-#endif
 }
