@@ -89,7 +89,8 @@ typedef struct _child_ {
  *	Global variables.
  */
 extern CHILD *family;
-extern int wrote_reboot;
+extern int wrote_wtmp_reboot;
+extern int wrote_utmp_reboot;
 
 /* Tokens in state parser */
 #define C_VER		1
@@ -107,7 +108,8 @@ extern int wrote_reboot;
 #define D_THISLEVEL    -3
 #define D_PREVLEVEL    -4
 #define D_GOTSIGN      -5
-#define D_WROTE_REBOOT -6
-#define D_SLTIME       -7
-#define D_DIDBOOT      -8
+#define D_WROTE_WTMP_REBOOT -6
+#define D_WROTE_UTMP_REBOOT -7
+#define D_SLTIME       -8
+#define D_DIDBOOT      -9
 
