@@ -6,10 +6,10 @@
  *
  * Author:	Miquel van Smoorenburg, miquels@cistron.nl
  *
- * Version:	@(#)last  2.78  24-Nov-1999  miquels@cistron.nl
+ * Version:	@(#)last  2.79  13-Jun-2001  miquels@cistron.nl
  *
  *		This file is part of the sysvinit suite,
- *		Copyright 1991-1999 Miquel van Smoorenburg.
+ *		Copyright 1991-2001 Miquel van Smoorenburg.
  *
  *		This program is free software; you can redistribute it and/or
  *		modify it under the terms of the GNU General Public License
@@ -40,7 +40,7 @@
 #  define SHUTDOWN_TIME 254
 #endif
 
-char *Version = "@(#) last 2.78 24-Nov-1999 miquels";
+char *Version = "@(#) last 2.79 11-Sep-2000 miquels";
 
 #define CHOP_DOMAIN	0	/* Define to chop off local domainname. */
 #define NEW_UTMP	1	/* Fancy & fast utmp read code. */
@@ -60,6 +60,8 @@ struct utmplist *utmplist = NULL;
 #define R_NORMAL	3 /* Normal */
 #define R_NOW		4 /* Still logged in */
 #define R_REBOOT	5 /* Reboot record. */
+#define R_PHANTOM	6 /* No logout record but session is stale. */
+#define R_TIMECHANGE	7 /* NEW_TIME or OLD_TIME */
 
 /* Global variables */
 int maxrecs = 0;	/* Maximum number of records to list. */
@@ -298,7 +300,7 @@ char *mybasename(char *s)
 /*
  *	Lookup a host with DNS.
  */
-int dns_lookup(char *result, char *org, unsigned int ip)
+int dns_lookup(char *result, int size, char *org, unsigned int ip)
 {
 	struct hostent *h;
 
@@ -306,17 +308,19 @@ int dns_lookup(char *result, char *org, unsigned int ip)
 	 *	Try to catch illegal IP numbers
 	 */
 	if (ip == 0 || (int)ip == -1 || (ip >> 24) == 0 || (ip & 255) == 0) {
-		strncpy(result, org, UT_HOSTSIZE);
-		result[UT_HOSTSIZE] = 0;
+		if (size > UT_HOSTSIZE) size = UT_HOSTSIZE+1;
+		strncpy(result, org, size - 1);
+		result[size-1] = 0;
 		return 0;
 	}
 
 	if ((h = gethostbyaddr((char *)&ip, 4, AF_INET)) == NULL) {
-		strcpy(result, inet_ntoa(*(struct in_addr *)&ip));
+		strncpy(result, inet_ntoa(*(struct in_addr *)&ip), size);
+		result[size-1] = 0;
 		return 0;
 	}
-	strncpy(result, h->h_name, 256);
-	result[255] = 0;
+	strncpy(result, h->h_name, size-1);
+	result[size-1] = 0;
 
 	return 0;
 }
@@ -389,9 +393,18 @@ int list(struct utmp *p, time_t t, int what)
 			sprintf(logouttime, "  still");
 			sprintf(length, "logged in");
 			break;
+		case R_PHANTOM:
+			length[0] = 0;
+			sprintf(logouttime, "   gone");
+			sprintf(length, "- no logout");
+			break;
 		case R_REBOOT:
 			logouttime[0] = 0;      /* Print machine uptime */
 			break;
+		case R_TIMECHANGE:
+			logouttime[0] = 0;
+			length[0] = 0;
+			break;
 		case R_NORMAL:
 			break;
  	}
@@ -400,7 +413,7 @@ int list(struct utmp *p, time_t t, int what)
 	 *	Look up host with DNS if needed.
 	 */
 	if (usedns)
-		dns_lookup(domain, p->ut_host, p->ut_addr);
+		dns_lookup(domain, sizeof(domain), p->ut_host, p->ut_addr);
 	if (useip) {
 		in.s_addr = p->ut_addr;
 		strcpy(domain, inet_ntoa(in));
@@ -418,17 +431,20 @@ int list(struct utmp *p, time_t t, int what)
 		     strcmp(s + 1, domainname) == 0) *s = 0;
 #endif
 		if (!altlist) {
-			sprintf(final, "%-8.8s %-12.12s %-16.16s %-16.16s %-7.7s %-12.12s\n",
+			snprintf(final, sizeof(final),
+				"%-8.8s %-12.12s %-16.16s "
+				"%-16.16s %-7.7s %-12.12s\n",
 				p->ut_name, utline,
 				domain, logintime, logouttime, length);
 		} else {
-			sprintf(final,
+			snprintf(final, sizeof(final), 
 				"%-8.8s %-12.12s %-16.16s %-7.7s %-12.12s %s\n",
 				p->ut_name, utline,
 				logintime, logouttime, length, domain);
 		}
 	} else
-		sprintf(final, "%-8.8s %-12.12s %-16.16s %-7.7s %-12.12s\n",
+		snprintf(final, sizeof(final),
+			"%-8.8s %-12.12s %-16.16s %-7.7s %-12.12s\n",
 			p->ut_name, utline,
 			logintime, logouttime, length);
 
@@ -436,7 +452,7 @@ int list(struct utmp *p, time_t t, int what)
 	 *	Print out "final" string safely.
 	 */
 	for (s = final; *s; s++) {
-		if (*s == '\n' || (*s >= 32 && (unsigned char)*s <= 128))
+		if (*s == '\n' || (*s >= 32 && (unsigned char)*s <= 126))
 			putchar(*s);
 		else
 			putchar('*');
@@ -546,11 +562,21 @@ int main(int argc, char **argv)
   lastdate = lastdown;
 
 #if CHOP_DOMAIN
-  /* Find out domainname. */
-  (void) gethostname(hostname, 256);
+  /*
+   *	Find out domainname.
+   *
+   *	This doesn't work on modern systems, where only a DNS
+   *	lookup of the result from hostname() will get you the domainname.
+   *	Remember that domainname() is the NIS domainname, not DNS.
+   *	So basically this whole piece of code is bullshit.
+   */
+  hostname[0] = 0;
+  (void) gethostname(hostname, sizeof(hostname));
   if ((domainname = strchr(hostname, '.')) != NULL) domainname++;
   if (domainname == NULL || domainname[0] == 0) {
-	(void) getdomainname(hostname, 256);
+	hostname[0] = 0;
+	(void) getdomainname(hostname, sizeof(hostname));
+	hostname[sizeof(hostname) - 1] = 0;
 	domainname = hostname;
 	if (strcmp(domainname, "(none)") == 0 || domainname[0] == 0)
 		domainname = NULL;
@@ -642,6 +668,14 @@ int main(int argc, char **argv)
 		 */
 		if (ut.ut_name[0] == 0)
 			ut.ut_type = DEAD_PROCESS;
+
+		/*
+		 *	Clock changes.
+		 */
+		if (strcmp(ut.ut_name, "date") == 0) {
+			if (ut.ut_line[0] == '|') ut.ut_type = OLD_TIME;
+			if (ut.ut_line[0] == '{') ut.ut_type = NEW_TIME;
+		}
 	}
 #endif
 
@@ -654,6 +688,15 @@ int main(int argc, char **argv)
 			lastdown = lastrch = ut.ut_time;
 			down = 1;
 			break;
+		case OLD_TIME:
+		case NEW_TIME:
+			if (extended) {
+				strcpy(ut.ut_line,
+				ut.ut_type == NEW_TIME ? "new time" :
+					"old time");
+				quit = list(&ut, lastdown, R_TIMECHANGE);
+			}
+			break;
 		case BOOT_TIME:
 			strcpy(ut.ut_line, "system boot");
 			quit = list(&ut, lastdown, R_REBOOT);
@@ -699,14 +742,20 @@ int main(int argc, char **argv)
 				}
 			}
 			/*
-			 *	Not found? Then crashed, down
-			 *	or still logged in
+			 *	Not found? Then crashed, down, still
+			 *	logged in, or missing logout record.
 			 */
 			if (c == 0) {
-				if (lastboot == 0)
-					quit = list(&ut, time(NULL), R_NOW);
-				else
-					quit = list(&ut, lastboot, whydown);
+				if (lastboot == 0) {
+					c = R_NOW;
+					/* Is process still alive? */
+					if (ut.ut_pid > 0 &&
+					    kill(ut.ut_pid, 0) != 0 &&
+					    errno == ESRCH)
+						c = R_PHANTOM;
+				} else
+					c = whydown;
+				quit = list(&ut, lastboot, c);
 			}
 			/* FALLTHRU */
 
