@@ -13,10 +13,10 @@
  *
  * Author:	Miquel van Smoorenburg, miquels@cistron.nl
  *
- * Version:	@(#)shutdown  2.78  13-Nov-1999  miquels@cistron.nl
+ * Version:	@(#)shutdown  2.83  05-Sep-2001  miquels@cistron.nl
  *
  *		This file is part of the sysvinit suite,
- *		Copyright 1991-1999 Miquel van Smoorenburg.
+ *		Copyright 1991-2001 Miquel van Smoorenburg.
  *
  *		This program is free software; you can redistribute it and/or
  *		modify it under the terms of the GNU General Public License
@@ -25,6 +25,7 @@
  */
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <sys/wait.h>
 #include <time.h>
 #include <string.h>
 #include <unistd.h>
@@ -33,11 +34,12 @@
 #include <stdio.h> 
 #include <signal.h>
 #include <fcntl.h>
+#include <stdarg.h>
 #include <utmp.h>
 #include "paths.h"
 #include "reboot.h"
 
-char *Version = "@(#) shutdown 2.78 13-Nov-1999 miquels@cistron.nl";
+char *Version = "@(#) shutdown 2.79 26-Feb-2001 miquels@cistron.nl";
 
 #define MESSAGELEN	256
 
@@ -51,6 +53,13 @@ char *sltime = 0;	/* Sleep time			*/
 char newstate[64];	/* What are we gonna do		*/
 int doself = 0;		/* Don't use init		*/
 
+char *clean_env[] = {
+	"HOME=/",
+	"PATH=/bin:/usr/bin:/sbin:/usr/sbin",
+	"TERM=dumb",
+	NULL,
+};
+
 /* From "wall.c" */
 extern void wall(char *, int, int);
 
@@ -110,18 +119,20 @@ void usage()
 void warn(mins)
 int mins;
 {
-  char buf[MESSAGELEN + 64];
+  char buf[MESSAGELEN + sizeof(newstate)];
   int len;
 
-  strcpy(buf, message);
+  strncpy(buf, message, sizeof(buf));
+  buf[sizeof(buf) - 1] = 0;
   len = strlen(buf);
 
   if (mins == 0)
-	sprintf(buf + len, "\rThe system is going down %s NOW !!\r\n",
+	snprintf(buf + len, sizeof(buf) - len,
+		"\rThe system is going down %s NOW!\r\n",
 		newstate);
   else
-  	sprintf(buf + len,
-		"\rThe system is going DOWN %s in %d minute%s !!\r\n",
+  	snprintf(buf + len, sizeof(buf) - len,
+		"\rThe system is going DOWN %s in %d minute%s!\r\n",
 			newstate, mins, mins == 1 ? "" : "s");
   wall(buf, 1, 0);
 }
@@ -145,6 +156,50 @@ void donologin(int min)
   }
 }
 
+/*
+ *	Spawn an external program.
+ */
+int spawn(int noerr, char *prog, ...)
+{
+	va_list	ap;
+	pid_t	pid;
+	int	i;
+	char	*argv[8];
+
+	i = 0;
+	while ((pid = fork()) < 0 && i < 10) {
+		perror("fork");
+		sleep(5);
+		i++;
+	}
+
+	if (pid < 0) return -1;
+
+	if (pid > 0) {
+		while(wait(&i) != pid)
+			;
+		return WEXITSTATUS(i);
+	}
+
+	if (noerr) fclose(stderr);
+
+	va_start(ap, prog);
+	for (i = 0; i < 7 && (argv[i] = va_arg(ap, char *)) != NULL; i++)
+		;
+	argv[i] = NULL;
+	va_end(ap);
+
+	chdir("/");
+	environ = clean_env;
+
+	execvp(argv[0], argv);
+	perror(argv[0]);
+	exit(1);
+
+	/*NOTREACHED*/
+	return 0;
+}
+
 /* Kill all processes, call /etc/init.d/halt (if present) */
 void fastdown()
 {
@@ -172,7 +227,11 @@ void fastdown()
 #endif
 
   /* First close all files. */
-  for(i = 0; i < 3; i++) if (!isatty(i)) close(i);
+  for(i = 0; i < 3; i++)
+	if (!isatty(i)) {
+		close(i);
+		open("/dev/null", O_RDWR);
+	}
   for(i = 3; i < 20; i++) close(i);
   close(255);
 
@@ -195,28 +254,26 @@ void fastdown()
 #if 0
   /* See if we can run /etc/init.d/halt */
   if (access(script, X_OK) == 0) {
-	sprintf(cmd, "%s fast", script);
-	system(cmd);
+	spawn(1, cmd, "fast", NULL);
 	fprintf(stderr, "shutdown: %s returned - falling back on default routines\r\n", script);
   }
 #endif
 
   /* script failed or not present: do it ourself. */
   sleep(1); /* Give init the chance to collect zombies. */
-  setenv("PATH", "/bin:/sbin:/usr/bin:/usr/sbin", 1);
 
   /* Record the fact that we're going down */
   write_wtmp("shutdown", "~~", 0, RUN_LVL, "~~");
 
   /* This is for those who have quota installed. */
-  system("accton 2>/dev/null");
-  system("quotaoff -a 2>/dev/null");
+  spawn(1, "accton", NULL);
+  spawn(1, "quotaoff", "-a", NULL);
 
   sync();
   fprintf(stderr, "shutdown: turning off swap\r\n");
-  system("swapoff -a");
+  spawn(0, "swapoff", "-a", NULL);
   fprintf(stderr, "shutdown: unmounting all file systems\r\n");
-  system("umount -a");
+  spawn(0, "umount", "-a", NULL);
 
   /* We're done, halt or reboot now. */
   if (do_halt) {
@@ -270,6 +327,25 @@ void shutdown()
   exit(1);
 }
 
+/*
+ *	returns if a warning is to be sent for wt
+ */
+static int needwarning(int wt)
+{
+	int ret;
+
+	if (wt < 10)
+		ret = 1;
+	else if (wt < 60)
+		ret = (wt % 15 == 0);
+	else if (wt < 180)
+		ret = (wt % 30 == 0);
+	else
+		ret = (wt % 60 == 0);
+
+	return ret;
+}
+
 /*
  * Main program.
  * Process the options and do the final countdown.
@@ -293,6 +369,7 @@ char **argv;
   FILE *fp;
   char *downusers[32];
   char buf[128];
+  char term[UT_LINESIZE + 6];
   struct stat st;
   struct utmp *ut;
   int user_ok = 0;
@@ -303,7 +380,7 @@ char **argv;
   setuid(geteuid());
 
   if (getuid() != 0) {
-  	fprintf(stderr, "shutdown: must be root.\n");
+  	fprintf(stderr, "shutdown: you must be root to do that!\n");
   	exit(1);
   }
   strcpy(down_level, "1");
@@ -377,10 +454,14 @@ char **argv;
 
 		/* See if this is a user process on a VC. */
 		if (ut->ut_type != USER_PROCESS) continue;
-		sprintf(buf, "/dev/%s", ut->ut_line);
-		if (stat(buf, &st) < 0) continue;
+		sprintf(term, "/dev/%.*s", UT_LINESIZE, ut->ut_line);
+		if (stat(term, &st) < 0) continue;
+#ifdef major /* glibc */
+		if (major(st.st_rdev) != 4 ||
+		    minor(st.st_rdev) > 63) continue;
+#else
 		if ((st.st_rdev & 0xFFC0) != 0x0400) continue;
-
+#endif
 		/* Root is always OK. */
 		if (strcmp(ut->ut_user, "root") == 0) {
 			user_ok++;
@@ -450,7 +531,7 @@ char **argv;
 
   /* Extra check. */
   if (doself && down_level[0] != '0' && down_level[0] != '6') {
-	fprintf(stderr, "shutdown: can only use \"-n\" for halt or reboot.\r\n");
+	fprintf(stderr, "shutdown: can use \"-n\" for halt or reboot only.\r\n");
 	exit(1);
   }
 
@@ -503,6 +584,10 @@ char **argv;
   if (when[0] == '+') when++;
 
   /* Decode shutdown time. */
+  for (sp = when; *sp; sp++) {
+	if (*sp != ':' && (*sp < '0' || *sp > '9'))
+		usage();
+  }
   if (strchr(when, ':') == NULL) {
 	/* Time in minutes. */
 	wt = atoi(when);
@@ -520,13 +605,13 @@ char **argv;
   if (wt == 0) shutdown();
 
   /* Give warnings on regular intervals and finally shutdown. */
-  if (wt < 15 && wt!= 10 && wt != 5 && wt != 1) warn(wt);
+  if (wt < 15 && !needwarning(wt)) warn(wt);
   while(wt) {
 	if (wt <= 5 && !didnolog) {
 		donologin(wt);
 		didnolog++;
 	}
-	if (wt == 15 || wt == 10 || wt == 5 || wt == 1) warn(wt);
+	if (needwarning(wt)) warn(wt);
 	hardsleep(60);
 	wt--;
   }
